@@ -4,8 +4,8 @@
  * This file is based on ECMA-167 3rd edition (June 1997)
  * http://www.ecma.ch
  *
- * Copyright (c) 2001-2002  Ben Fennema <bfennema@falcon.csc.calpoly.edu>
- * Copyright (c) 2017       Pali Rohár <pali.rohar@gmail.com>
+ * Copyright (c) 2001-2002  Ben Fennema
+ * Copyright (c) 2017-2018  Pali Rohár <pali.rohar@gmail.com>
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -38,7 +38,7 @@
  * ECMA-167r3 defines and structure definitions
  */
 
-#include <inttypes.h>
+#include <stdint.h>
 
 #ifndef _ECMA_167_H
 #define _ECMA_167_H 1
@@ -746,6 +746,7 @@ struct appUseExtAttr
 #define EXTATTR_DEV_SPEC		12
 #define EXTATTR_IMP_USE			2048
 #define EXTATTR_APP_USE			65536
+#define EXTATTR_SUBTYPE			1
 
 
 /* Unallocated Space Entry (ECMA 167r3 4/14.11) */
