@@ -1097,12 +1097,17 @@ int
 questionOverwrite(Directory *dir, struct fileIdentDesc *fid, char* name)
 {
     printf("File %s already exists. Overwrite ? (y/N) : ", name);
-#ifdef _GNU_SOURCE
+#ifdef USE_READLINE
     readLine(NULL);
 #else
-    fgets(line, 256, stdin);
+    if (!fgets(line, 256, stdin))
+	line[0] = 0;
 #endif
-    if( !line || line[0] != 'y' )
+#ifdef USE_READLINE
+    if( !line )
+	return 1;
+#endif
+    if( line[0] != 'y' )
 	return 1;
     deleteFID(dir, fid);
     return 0;
